
/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 *
 * Copyright 1997-2007 Sun Microsystems, Inc. All rights reserved.
 *
 * The contents of this file are subject to the terms of either the GNU
 * General Public License Version 2 only ("GPL") or the Common
 * Development and Distribution License("CDDL") (collectively, the
 * "License"). You may not use this file except in compliance with the
 * License. You can obtain a copy of the License at
 * http://www.netbeans.org/cddl-gplv2.html
 * or nbbuild/licenses/CDDL-GPL-2-CP. See the License for the
 * specific language governing permissions and limitations under the
 * License.  When distributing the software, include this License Header
 * Notice in each file and include the License file at
 * nbbuild/licenses/CDDL-GPL-2-CP.  Sun designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Sun in the GPL Version 2 section of the License file that
 * accompanied this code. If applicable, add the following below the
 * License Header, with the fields enclosed by brackets [] replaced by
 * your own identifying information:
 * "Portions Copyrighted [year] [name of copyright owner]"
 *
 * Contributor(s):
 *
 * The Original Software is NetBeans. The Initial Developer of the Original
 * Software is Sun Microsystems, Inc. Portions Copyright 1997-2006 Sun
 * Microsystems, Inc. All Rights Reserved.
 *
 * If you wish your version of this file to be governed by only the CDDL
 * or only the GPL Version 2, indicate your decision by adding
 * "[Contributor] elects to include this software in this distribution
 * under the [CDDL or GPL Version 2] license." If you do not indicate a
 * single choice of license, a recipient has the option to distribute
 * your version of this file under either the CDDL, the GPL Version 2 or
 * to extend the choice of license to its licensees as provided above.
 * However, if you add GPL Version 2 code and therefore, elected the GPL
 * Version 2 license, then the option applies only if the new code is
 * made subject to such option by the copyright holder.
 */
package org.netbeans.modules.coverage.annotations;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.util.HashMap;
import java.util.Iterator;
import org.openide.text.Annotatable;
import org.openide.text.Annotation;
import org.openide.text.Line;

/**
 * Base annotation class that contains some basic logic and holds information about
 *all annotations for all classes.
 * @author Andrey Korostelev
 **/
public abstract class AbstractCoverageAnnotation extends Annotation implements PropertyChangeListener {
    
    private static HashMap<String, HashMap<Integer, Annotation>> annotations = new HashMap<String, HashMap<Integer, Annotation>>();
    private String descrmessage = null;
    private final static String KEY_JOIN_CHAR = " ";
    
    
    
    /**
     * Creates new annotation and adds it to the global annotations list so it
     * can be managed later
     * @param projectName Project this annotation belongs to
     * @param className Class this annotation belongs to
     * @param lineNum Number of the line to be annotated
     */
    public AbstractCoverageAnnotation(String projectName, String className, Integer lineNum){
        synchronized(annotations) {
            String key = combineKey(projectName, className);
            HashMap<Integer, Annotation> anns = annotations.get(key);
            if (anns==null){
                anns = new HashMap<Integer, Annotation>();
                annotations.put(key, anns);
            }
            anns.put(lineNum, this);
        }
    }
    
    /**
     * Removes all editor annotations from specified project.
     * <br>
     * Currently uses slow string comparing approach to select annotations maps for current project.
     * Might be changed to use more complicated storage structure, 
     * but it will make annotations storing more difficult.
     * @param projectName Project to remove annotations for
     */
    public static void removeAll(String projectName){
        synchronized(annotations) {
            for (Iterator<String> it = annotations.keySet().iterator(); it.hasNext();){
                String key = it.next();
                if (key.startsWith(projectName+KEY_JOIN_CHAR)){
                    HashMap<Integer, Annotation> anns = annotations.get(key);
                    for (Iterator<Annotation> lit = anns.values().iterator(); lit.hasNext();){
                        lit.next().detach();
                    }
                    anns.clear();
                    it.remove();
                }
            }
        }
    }
    
    /**
     * Removes annotations from specific class in the specific project
     * @param projectName Project to remove annotations for
     * @param className Class to remove annotations for
     */
    public static void removeFromClass(String projectName, String className){
        String id = combineKey(projectName, className);
        synchronized(annotations) {
            HashMap<Integer, Annotation> anns = annotations.get(id);
            if (anns!=null){
                for (Iterator<Annotation> lit = anns.values().iterator(); lit.hasNext();){
                    lit.next().detach();
                }
                anns.clear();
                annotations.remove(id);
            }
        }
    }
    
       
    
    /**
     * Sets the current descrmessage
     *
     * @param message the descrmessage
     */
    public void setMessage( String message ) {
        descrmessage = message;
    }
    
    /**
     * Getter for annotations descr. message
     * @return message
     */
    public String getMessage() {
        return descrmessage;
    }
    
    /**
     * implemented  from PropertyChangeListener.
     * Detaches itself as annotation and property change listener
     * on all actions except
     * Annotatable.PROP_ANNOTATION_COUNT (e.g.: adding/removing breakpoint)
     */
    public void propertyChange( PropertyChangeEvent propertyChangeEvent ) {
        if (propertyChangeEvent.getPropertyName() != Annotatable.PROP_ANNOTATION_COUNT){
            Line line = ( Line )propertyChangeEvent.getSource();
            line.removePropertyChangeListener( this );
            detach();
            annotations.remove(this);
        }
    }
    
    /**
     *This method is used to create a unique key for every new class' annotation
     **/
    private static String combineKey(String projectName, String className){
        return projectName+KEY_JOIN_CHAR+className;
    }
    
}
