/*
 * The contents of this file are subject to the terms of the Common Development
 * and Distribution License (the License). You may not use this file except in
 * compliance with the License.
 * 
 * You can obtain a copy of the License at http://www.netbeans.org/cddl.html
 * or http://www.netbeans.org/cddl.txt.
 * 
 * When distributing Covered Code, include this CDDL Header Notice in each file
 * and include the License file at http://www.netbeans.org/cddl.txt.
 * If applicable, add the following below the CDDL Header, with the fields
 * enclosed by brackets [] replaced by your own identifying information:
 * "Portions Copyrighted [year] [name of copyright owner]"
 * 
 * The Original Software is NetBeans. The Initial Developer of the Original
 * Software is Sun Microsystems, Inc. Portions Copyright 1997-2007 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

/*
 * ConnectionOperator.java
 */

package org.netbeans.orchlib.core.operators;

import java.awt.Component;
import java.util.Iterator;
import java.util.List;
import org.netbeans.orchlib.exceptions.ElementNotFoundException;
import org.netbeans.jemmy.JemmyProperties;
import org.netbeans.jemmy.Timeouts;
import org.netbeans.jemmy.Waitable;
import org.netbeans.jemmy.Waiter;
import org.netbeans.jemmy.operators.Operator;
import org.netbeans.modules.bpel.design.model.connections.Connection;
import org.netbeans.modules.bpel.design.model.elements.VisualElement;
import org.netbeans.modules.bpel.design.model.patterns.CompositePattern;



/**
 * The class represents a connection between two visual elements and provides some basic
 * methods to analyze, test the connections.
 *
 * Please be aware that connections can not be operated through UI in any way so
 * this class is mostly used for internal purposes and to check connections existance
 *
 * @author Mikhail.Kondratyev@sun.com
 */
public class ConnectionOperator  extends Operator {
    
    private Connection connection = null;
    
    
    public final static long WAIT_CONNECTION_TIMEOUT = 60000;
    
    
    public ConnectionOperator(VisualElementOperator source, VisualElementOperator target) throws ElementNotFoundException {
        this(waitForConnection(source, target));
    }
    
    
    public ConnectionOperator(Connection connection) {
        this.connection = connection;
    }
    
    public ConnectionOperator(String sourceName, String targetName, PatternOperator commonContainer) throws ElementNotFoundException {
        this(sourceName, 0, targetName, 0,  commonContainer);
    }
    
    
    
    public ConnectionOperator(String sourceName, int sourceIndex, String targetName, int targetIndex, PatternOperator commonContainer) throws ElementNotFoundException {
        this(waitForConnection(
                new VisualElementOperator(commonContainer, sourceName, sourceIndex),
                new VisualElementOperator(commonContainer, targetName, targetIndex))
                );
    }
    
    
    private static Connection findConnection(VisualElementOperator source, VisualElementOperator target) throws ElementNotFoundException{
        
        List<Connection> outConns =  source.getVisualElement().getOutcomingConnections();
                
        for (Iterator<Connection> iter = outConns.iterator(); iter.hasNext();) {
            Connection connOut = iter.next();
            VisualElement brdElement = null;
            VisualElement trgElement = target.getVisualElement();
            if(trgElement.getPattern() instanceof CompositePattern){
                brdElement = ((CompositePattern)trgElement.getPattern()).getBorder();
            }
        
           if (connOut.getSource()==source.getVisualElement() && (connOut.getTarget()==trgElement || connOut.getTarget()==brdElement)){
                return connOut;
            }            
        }
        
        throw new ElementNotFoundException();        
    }
        
    
    
    public static Connection waitForConnection(final VisualElementOperator source, final VisualElementOperator target) {
        try{
            Waiter w = new Waiter(new Waitable() {
                public Object actionProduced(Object obj) {
                    try {
                        Connection go = findConnection(source, target);
                        return go;
                    }catch(ElementNotFoundException e) {
                        return null;
                    }
                }
                public String getDescription() {
                    return("Wait for connection between elems" );
                }
            });
            Timeouts t = JemmyProperties.getCurrentTimeouts();
            t.setTimeout("Waiter.WaitingTime", t.getTimeout("ConnectionOperator.WaitConnectionTime"));
            return (Connection)w.waitAction(null);
            
        }catch(InterruptedException ie) {
            return null;
        }
    }
    
    
    
    /**
     * We must implement this method to extend Operator, but since
     * we are not working with Component this method always return null;
     * @return null
     */
    public Component getSource() {
        return null;
    }
    
    Connection getConnection(){
        return connection;
    }
    
    public VisualElementOperator getSourceElement(){
        return new VisualElementOperator(connection.getSource());
    }
    
    public VisualElementOperator getTargetElement(){
        return new VisualElementOperator(connection.getTarget());
    }
       
    
    public boolean equals(Object el){
        if (el instanceof ConnectionOperator){
            return (this.connection == ((ConnectionOperator)el).connection);
        }else{
            return false;
        }
    }
    
    static {
        Timeouts.initDefault("ConnectionOperator.WaitConnectionTime", WAIT_CONNECTION_TIMEOUT);
    }
    
}


