/*
 * The contents of this file are subject to the terms of the Common Development
 * and Distribution License (the License). You may not use this file except in
 * compliance with the License.
 *
 * You can obtain a copy of the License at http://www.netbeans.org/cddl.html
 * or http://www.netbeans.org/cddl.txt.
 *
 * When distributing Covered Code, include this CDDL Header Notice in each file
 * and include the License file at http://www.netbeans.org/cddl.txt.
 * If applicable, add the following below the CDDL Header, with the fields
 * enclosed by brackets [] replaced by your own identifying information:
 * "Portions Copyrighted [year] [name of copyright owner]"
 *
 * The Original Software is NetBeans. The Initial Developer of the Original
 * Software is Sun Microsystems, Inc. Portions Copyright 1997-2007 Sun
 * Microsystems, Inc. All Rights Reserved.
 */


/*
 * DiagramOperator.java
 *
 */

package org.netbeans.orchlib.core.operators;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.event.KeyEvent;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import org.netbeans.orchlib.exceptions.ElementNotFoundException;
import org.netbeans.jellytools.TopComponentOperator;
import org.netbeans.jemmy.EventTool;
import org.netbeans.jemmy.Timeout;
import org.netbeans.jemmy.drivers.input.MouseRobotDriver;
import org.netbeans.jemmy.operators.JPopupMenuOperator;
import org.netbeans.modules.bpel.design.DesignView;
import org.netbeans.modules.bpel.design.geometry.FPath;
import org.netbeans.modules.bpel.design.geometry.FPoint;
import org.netbeans.modules.bpel.design.geometry.FRectangle;
import org.netbeans.modules.bpel.design.model.patterns.CompositePattern;
import org.netbeans.modules.bpel.design.model.patterns.Pattern;
import org.netbeans.modules.bpel.editors.multiview.DesignerMultiViewElement;

/**
 * This is a very important class that allows to test  actions available for
 * the BPEL diagram. The samples of such actions are clicking mouse, creating new BPEL
 * elements, scrolling diagram, etc...
 * @author Mikhail.Kondratyev@sun.com
 */
public class DesignViewOperator extends TopComponentOperator{
    
    //the width of border we will use
    private final int WIDTH_BORDER = 100;
    private final int HEIGHT_BORDER = 100;
    
    private DesignView designView = null;
    
    static EventTool m_eventTool = new EventTool();
    
    /** Creates a new instance of DiagramOperator */
    public DesignViewOperator(DesignerMultiViewElement comp) {
        super(comp.getDesignView().getView());
        designView = comp.getDesignView();
    }
    
    public DesignViewOperator(DesignView view) {
        super(view.getView());
        designView = view;
    }
    
    
    /**
     * Returns underlying DesignView
     * @return underlying DesignView element. The element contains
     * information about the elements present on the diagram
     */
    DesignView getDesignView(){
        return designView;
    }
    
    
    
    /**
     * Return free point.
     * You may use such point to place an element or invoke popup menu     *
     * @return free point
     */
    public Point getFreePoint() {
        int width = (int)getDesignView().getView().getWidth();
        int height = (int)getDesignView().getView().getHeight();
        DesignView view = getDesignView();
        for(int w=WIDTH_BORDER;w<width-WIDTH_BORDER;w+=10) {
            for(int h=HEIGHT_BORDER;h<height-HEIGHT_BORDER;h+=10 ) {
                Point p = new Point(w,h);
                Object obj = view.findElement((float)w, (float)h);
                if (obj == null){
                    scrollToPoint(p);
                    return p;
                }
            }
        }
        //no free point found
        Point p = new Point(width+100,height+100);
        scrollToPoint(p);
        return null;
    }
    
    
    
    /**
     * Checks if the specified point is free.
     * Free means there is no top level element covering this point
     * @return free point
     * @param point Point to check
     */
    public boolean isFreePoint(Point point){
        DesignView view = getDesignView();
        CompositePattern pattern = (CompositePattern)view.getModel().getRootPattern();
        Collection<Pattern> patterns = pattern.getNestedPatterns();
        ArrayList<Rectangle> rects = new ArrayList<Rectangle>();
        for (Iterator<Pattern> entries = patterns.iterator(); entries.hasNext();){
            Pattern child = entries.next();
            if (child instanceof CompositePattern){
                rects.add(convertRectToScreen((FRectangle) ((CompositePattern)child).getBorder().getBounds()));
            }
        }
        
        for(int i=0;i<rects.size();i++){
            if (rects.get(i).contains(point)){
                return false;
            }
        }
        return true;
    }
    
    /**
     * Returns the root pattern in the hierarchy
     * @return The root pattern
     */
    public PatternOperator getRootPatternOperator(){
        return new PatternOperator(this, this.getDesignView().getModel().getRootPattern());
    }
    
    
    /**
     * Scrolls to the point specified
     * @point point to scroll to
     * @param point Point to scroll to
     */
    public void scrollToPoint(Point point) {
        int margin = 75;
        int x = 0;
        int y = 0;
        if (point.x>margin){
            x = point.x-margin;
        }
        if (point.y>margin){
            y = point.y-margin;
        }
        this.scrollRectToVisible(new Rectangle(x, y, margin+margin, margin+margin));
    }
    
    
    public void select(){
        getOutput().printTrace("Diagram is not a selectable object.");
    }
    
    public double getZoom(){
        return getDesignView().getZoom();
    }
    
    public void addToSelection(){
        getOutput().printTrace("Diagram is not a selectable object.");
    }
    
    
    /**
     * Invokes popup menu
     * @return  popup menu operator
     */
    public JPopupMenuOperator getPopup(){
        Point clickPoint = getFreePoint();
        clickForPopup(clickPoint.x, clickPoint.y);
        return new JPopupMenuOperator();
    }
    
    /**
     * Invokes popup menu
     * @return  popup menu operator
     */
    public JPopupMenuOperator getPopup(Point p){
        clickForPopup(p.x, p.y);
        return new JPopupMenuOperator();
    }
    
    /**
     * Drags an element to the specified point on the diagram
     * @param element Actionable element that will be dragged
     * @param target Point to drag element to
     */
    public void dragElement(IActionable element, Point target){
        Point actionPoint = element.getActionPoint();
        
        scrollToPoint(actionPoint);
        moveMouse(actionPoint.x,  actionPoint.y);
        pressMouse(actionPoint.x, actionPoint.y);
        //TODO: need to add waitSelected here somehow
        
        scrollToPoint(target);
        moveMouse(target.x, target.y);
        try{
            Thread.sleep(1000);
        }catch(Exception e){}
        releaseMouse(target.x, target.y);
    }
    
    
    /**
     * Drags a specified pattern inside another pattern. BPEL
     * designer allows to drop patterns only at the specified
     * points on connection. The exact point is calculated automatically
     * by the method
     * @param pattern Pattern to drag
     * @param targetConnection Connection to drop pattern onto
     * @param targetPattern Target pattern
     */
    public void dragPattern(PatternOperator pattern, ConnectionOperator targetConnection, PatternOperator targetPattern){

        Point actionPoint = pattern.getActionPoint();
        Pattern ptrn = null;
        if (targetPattern.getPattern() instanceof CompositePattern){
            ptrn = targetPattern.getPattern();
        }else{
            //TODO: change this
            System.err.println("ERROR: dragPattern was called for Non-Composite target pattern. Method aborted");
            return;
        }
        
        pattern.makeVisible();
        moveMouse(actionPoint.x,  actionPoint.y);
        pressMouse(actionPoint.x, actionPoint.y);
        //TODO: need to add waitSelected here somehow
//        eventTool.waitNoEvent(500);
        try{
            Thread.sleep(1000);
        }catch(Exception e){}
        
        FPath connectedPath = targetConnection.getConnection().getSegmentsForPattern((CompositePattern)targetPattern.getPattern());
        if (connectedPath==null){
            //TODO in fact we should never come here
            return;
        }
        
        //FPath path = connectedPath;
        //Point target = convertPointToScreen(path.getBounds().center());
        //Point target = new Point((int)connectedPath.getBounds().getCenterX(), (int)connectedPath.getBounds().getCenterY());
        
        Point target = convertPointToScreen(new FPoint(connectedPath.getBounds().getCenterX(), connectedPath.getBounds().getCenterY()));
        
        //Point scrollPoint = new Point(target.x+50, target.y+50);
        scrollToPoint(target);
        
        moveMouse(target.x+5, target.y+5);
        try{
            Thread.sleep(500);
        }catch(Exception e){}
        moveMouse(target.x, target.y);
        try{
            Thread.sleep(2000);
        }catch(Exception e){}
        releaseMouse(target.x, target.y);
        
        m_eventTool.waitNoEvent(500);
    }
    
    
    /**
     * Method allows to drag a pattern to the specified placeholder
     * @param pattern Pattern to drag
     * @param placeholderElement Placeholder to drag pattern to
     * @throws orchlib.exceptions.ElementNotFoundException
     */
    public void dragPattern(PatternOperator pattern, VisualElementOperator placeholderElement) throws ElementNotFoundException{
        
        pattern.makeVisible();
        Point actionPoint = pattern.getActionPoint();
        
        moveMouse(actionPoint.x,  actionPoint.y);
        pressMouse(actionPoint.x, actionPoint.y);
        
        moveMouse(actionPoint.x + 5,  actionPoint.y + 5);
        
        try{
            Thread.sleep(1000);
        }catch(Exception e){}
        
        Rectangle r = placeholderElement.getBoundingRect();
        r.x -= 25;
        r.y -= 25;
        r.width += 50;
        r.height += 50;
        
        scrollRectToVisible(r);
        m_eventTool.waitNoEvent(500);
        
        r = placeholderElement.getBoundingRect();
        int targetX = r.x + r.width/2;
        int targetY = r.y + r.height/2;
        
        dragMouse(targetX + 5, targetY + 5);
        try{
            Thread.sleep(500);
        }catch(Exception e){}
        
        dragMouse(targetX, targetY);
        try{
            Thread.sleep(1000);
        }catch(Exception e){}
        
        releaseMouse(targetX, targetY);
        
        m_eventTool.waitNoEvent(500);
    }
    
    /**
     * Creates new activity by dragging it from palette to the
     * specified location on the diagram
     * @param patternType Type of the pattern to be created
     * @param targetConnection Connection to drop pattern at
     * @param targetPattern Pattern to drop pattern at
     * @throws orchlib.exceptions.ElementNotFoundException
     */
    public void createPatternFromPalette(PatternTypes patternType, ConnectionOperator targetConnection, PatternOperator targetPattern) throws ElementNotFoundException{

        PaletteOperator palette = new PaletteOperator();
        
        m_eventTool.waitNoEvent(500);
        
        FPath connectedPath = targetConnection.getConnection().getSegmentsForPattern((CompositePattern)targetPattern.getPattern());
        if (connectedPath==null ){
            //TODO in fact we should never come here
            return;
        }
        //FPath path = connectedPathes[0];
        //Point target = convertPointToScreen(path.center());
        
        Point target = convertPointToScreen(new FPoint(connectedPath.getBounds().getCenterX(), connectedPath.getBounds().getCenterY()));
        
        
        Point scrollPoint = new Point(target.x+50, target.y+50);
        scrollToPoint(scrollPoint);
/*
        AWTEventListener l = new AWTEventListener() {
            public void eventDispatched(AWTEvent event) {
                System.err.println("Event:" + event);
            }
        };
        Toolkit.getDefaultToolkit().addAWTEventListener(l ,0xFFFFFFFF);
 */
        //clicking on palette
        Point point = palette.pressMouseOnItem(patternType);
        MouseRobotDriver drv = new MouseRobotDriver(new Timeout("",  50));
        drv.exitMouse(palette);
        drv.enterMouse(this);
//        eventTool.waitNoEvent(500);
        
        try{
            Thread.sleep(500);
        }catch(Exception e){}
        
        drv.moveMouse(this, target.x, target.y);
        
//	eventTool.waitNoEvent(1000);
        
        try{
            Thread.sleep(500);
        }catch(Exception e){}
        drv.releaseMouse(this, target.x, target.y, KeyEvent.BUTTON1_MASK, 0);
//		Toolkit.getDefaultToolkit().removeAWTEventListener(l);
        m_eventTool.waitNoEvent(500);
    }
    
    //TODO: add this logic to previous method to handle situation in general
    /**
     * Creates new pattern by dragging it from palette and dropping
     * over some placeholder element
     */
    public void createPatternFromPalette(PatternTypes patternType, VisualElementOperator placeholderElement) throws ElementNotFoundException{

        PaletteOperator palette = new PaletteOperator();
        
        m_eventTool.waitNoEvent(500);
        
        Point target = new Point((int)placeholderElement.getBoundingRect().getCenterX(), (int)placeholderElement.getBoundingRect().getCenterY());
        
        //Point scrollPoint = new Point(target.x+50, target.y+50);
        //scrollToPoint(scrollPoint);
        scrollToPoint(target);
        
        //clicking on palette
        Point point = palette.pressMouseOnItem(patternType);
        MouseRobotDriver drv = new MouseRobotDriver(new Timeout("",  50));
        drv.exitMouse(palette);
        drv.enterMouse(this);
//        eventTool.waitNoEvent(500);
        try{
            Thread.sleep(1000);
        }catch(Exception e){}
        
        drv.moveMouse(this, target.x, target.y);
//        eventTool.waitNoEvent(1000);
        try{
            Thread.sleep(1000);
        }catch(Exception e){}
        
        drv.releaseMouse(this, target.x, target.y, KeyEvent.BUTTON1_MASK, 0);
        
        m_eventTool.waitNoEvent(500);
    }
    
    
    /**
     * Drags pattern from palette to some point on the diagram
     * @param patternType Type of pattern to be created
     * @param target Target point
     * @throws orchlib.exceptions.ElementNotFoundException
     */
    public void createPatternFromPalette(PatternTypes patternType, Point target) throws ElementNotFoundException{

        PaletteOperator palette = new PaletteOperator();
        
        m_eventTool.waitNoEvent(500);
        
        scrollToPoint(target);
        
        //clicking on palette
        Point point = palette.pressMouseOnItem(patternType);
        MouseRobotDriver drv = new MouseRobotDriver(new Timeout("",  50));
        drv.exitMouse(palette);
        drv.enterMouse(this);
//        eventTool.waitNoEvent(500);
        try{
            Thread.sleep(1000);
        }catch(Exception e){}
        drv.moveMouse(this, target.x, target.y);
//        eventTool.waitNoEvent(1000);
        try{
            Thread.sleep(1000);
        }catch(Exception e){}
        drv.releaseMouse(this, target.x, target.y, KeyEvent.BUTTON1_MASK, 0);
        
        m_eventTool.waitNoEvent(500);
    }
    
    
    /**
     * Allows to convert point from the internal coordinates to the screen ones
     */
    public Point convertPointToScreen(FPoint point) {
        return getDesignView().convertDiagramToScreen(point, getDesignView().getCorrectedZoom());
    }
    
    /**
     * Converts rectangle from the internal coordinates to the screen ones
     */
    public Rectangle convertRectToScreen(FRectangle fRectangle){
        Point point = convertPointToScreen(fRectangle.getTopLeft());
        return new Rectangle(point.x, point.y, (int)(fRectangle.width * designView.getCorrectedZoom()), (int)(fRectangle.height * designView.getCorrectedZoom()));
    }
    
    
}
