/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 *
 * Copyright 1997-2007 Sun Microsystems, Inc. All rights reserved.
 *
 * The contents of this file are subject to the terms of either the GNU
 * General Public License Version 2 only ("GPL") or the Common
 * Development and Distribution License("CDDL") (collectively, the
 * "License"). You may not use this file except in compliance with the
 * License. You can obtain a copy of the License at
 * http://www.netbeans.org/cddl-gplv2.html
 * or nbbuild/licenses/CDDL-GPL-2-CP. See the License for the
 * specific language governing permissions and limitations under the
 * License.  When distributing the software, include this License Header
 * Notice in each file and include the License file at
 * nbbuild/licenses/CDDL-GPL-2-CP.  Sun designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Sun in the GPL Version 2 section of the License file that
 * accompanied this code. If applicable, add the following below the
 * License Header, with the fields enclosed by brackets [] replaced by
 * your own identifying information:
 * "Portions Copyrighted [year] [name of copyright owner]"
 *
 * Contributor(s):
 *
 * The Original Software is NetBeans. The Initial Developer of the Original
 * Software is Sun Microsystems, Inc. Portions Copyright 1997-2006 Sun
 * Microsystems, Inc. All Rights Reserved.
 *
 * If you wish your version of this file to be governed by only the CDDL
 * or only the GPL Version 2, indicate your decision by adding
 * "[Contributor] elects to include this software in this distribution
 * under the [CDDL or GPL Version 2] license." If you do not indicate a
 * single choice of license, a recipient has the option to distribute
 * your version of this file under either the CDDL, the GPL Version 2 or
 * to extend the choice of license to its licensees as provided above.
 * However, if you add GPL Version 2 code and therefore, elected the GPL
 * Version 2 license, then the option applies only if the new code is
 * made subject to such option by the copyright holder.
 */

package org.netbeans.server.uihandler;

import java.io.BufferedOutputStream;
import java.io.BufferedReader;
import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.UnsupportedEncodingException;
import java.security.Principal;
import java.util.Collections;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Locale;
import java.util.Map;
import java.util.concurrent.ExecutionException;
import java.util.logging.Level;
import javax.servlet.RequestDispatcher;
import javax.servlet.ServletContext;
import javax.servlet.ServletInputStream;
import javax.servlet.http.Cookie;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpSession;
import javax.servlet.http.HttpSessionContext;
import javax.servlet.jsp.PageContext;
import org.netbeans.junit.NbTestCase;
import org.netbeans.lib.uihandler.InputGesture;
import org.netbeans.modules.exceptions.entity.Logfile;
import org.netbeans.modules.exceptions.utils.PersistenceUtils;
import org.netbeans.server.uihandler.StatisticsTest.PgCnt;

/**
 *
 * @author Jaroslav Tulach
 */
public class LogsManagerTest extends NbTestCase {
    private static final String LOCALHOST_IP ="127.0.0.1";
    
    private LogsManager result;
    private File logs;
    
    static {
        System.setProperty("uihandlerserver.timeout", "10000");
    }
    
    
    public LogsManagerTest(String testName) {
        super(testName);
    }
    
    @Override
    protected Level logLevel() {
        return Level.FINE;
    }
    
    private File logs() throws IOException {
        File f = new File(getWorkDir(), "logs");
        f.mkdirs();
        return f;
    }
    
    
    protected void setUp() throws Exception {
        clearWorkDir();
        DatabaseTestCase.initPersitenceUtils(getWorkDir());
        
        File toDir = logs();
        File log = extractResourceAs(toDir, "1.log", "log444");
        File log2 = extractResourceAs(toDir, "2.log", "log4442");
        
        result = LogsManager.createManager(logs());
    }
    
    protected void tearDown() throws Exception {
    }
    
    public void testErrorWhenNoAnnotationGiven() throws Exception {
        PageContext context = createPageContext();
        try {
            result.preparePageContext(context, Collections.singleton("UnknownStatistic"));
            fail("Should report unknown statistic");
        } catch (IllegalArgumentException ex) {
            if (ex.getMessage().indexOf("UnknownStatistic") == -1) {
                fail("Report message shall contain info about UnknownStatistic:\n" + ex.getMessage());
            }
        }
    }
    
    public void testIWantLogfileForAllLogFiles() throws Exception {
        String logName = "log444";
        Logfile log = result.getLogInfo(logName);
        assertNotNull("There is Logfile for every log", log);
        assertEquals(log.getUserdir(), logName);
        assertEquals(log.getUploadNumber(), new Integer(0));
    }

    public void testGetParsedLogs() throws Exception {
        result.preparePageContext(createPageContext(), "unknown", null);
        assertEquals("All parsed", result.getNumberOfLogs(), result.getParsedLogs());
        assertEquals("Two parsed", 2, result.getParsedLogs());

        File log3 = extractResourceAs(logs(), "2.log", "log444.1");
        
        result.addLog(log3, "127.0.0.1");
        result.preparePageContext(createPageContext(), "unknown", null);

        assertEquals("All 3 parsed", result.getNumberOfLogs(), result.getParsedLogs());
        assertEquals("Three parsed", 3, result.getParsedLogs());
    }
    
    public void testGetDefault() throws Exception {
        doCountsTest(0);
    }
    public void testUserInc() throws Exception {
        doCountsTest(1);
    }
    public void testUserSessionChange() throws Exception {
        doCountsTest(2);
    }
    
    private void doCountsTest(int type) throws Exception {
        assertNotNull(result);
        
        
        Map<InputGesture,Integer> cnts = null;
        switch (type) {
        case 0: cnts = getGlobalCounts(result); break;
        case 1: cnts = getUserCounts(result, "log444"); break;
        case 2: cnts = Collections.emptyMap();
        }
        
        assertNotNull(cnts);
        
        for (Map.Entry<InputGesture,Integer> e : cnts.entrySet()) {
            if (e.getValue() <= 0) {
                fail("Too less of " + e.getKey() + " = " + e.getValue());
            }
        }
        int numb = result.getNumberOfLogs();
        
        File log = extractResourceAs(logs(), "1.log", "log444.1");
        result.addLog(log, "10.0.0.23");
        
        Map<InputGesture,Integer> newCnts = null;
        switch (type) {
        case 0: newCnts = getGlobalCounts(result); break;
        case 1: newCnts = getUserCounts(result, "log444"); break;
        case 2: newCnts = getLastSessionCounts(result, "log444"); break;
        }
        assertEquals("One log added", numb + 1, result.getNumberOfLogs());
        
        for (Map.Entry<InputGesture,Integer> e : cnts.entrySet()) {
            Integer newValue = newCnts.get(e.getKey());
            if (e.getValue() >= newValue) {
                fail("Too small increment of of " + e.getKey() + " = " + e.getValue() + " and " + newValue);
            }
        }
        
        
        Logfile info = result.getLogInfo(log.getName());
        assertNotNull("There is a log info", info);
        assertEquals("address is ok", info.getIpAddr(), "10.0.0.23");
    }
    
    static File extractResource(File toDir, String res) throws IOException {
        return extractResourceAs(toDir, res, res);
    }
    
    static File extractResourceAs(File toDir, String res, String outFile) throws IOException {
        return extractResourceAs(LogsManagerTest.class, toDir, res, outFile);
    }
    public static File extractResourceAs(Class<?> caller, File toDir, String res, String outFile) throws IOException {
        InputStream is = caller.getResourceAsStream(res);
        File out = new File(toDir, outFile);
        OutputStream os = new BufferedOutputStream(new FileOutputStream(out));
        byte[] arr = new byte[4096];
        for (;;) {
            int len = is.read(arr);
            if (-1 == len) break;
            os.write(arr, 0, len);
        }
        os.close();
        is.close();
        return out;
    }

    public static File extractString(File toDir, String rec, String outFile) throws IOException {
        InputStream is = new ByteArrayInputStream(rec.getBytes());
        File out = new File(toDir, outFile);
        OutputStream os = new BufferedOutputStream(new FileOutputStream(out));
        byte[] arr = new byte[4096];
        for (;;) {
            int len = is.read(arr);
            if (-1 == len) break;
            os.write(arr, 0, len);
        }
        os.close();
        is.close();
        return out;
    }

    public void testGetIP(){
        Map<String, String> header = new HashMap<String, String>();
        TestHttpSevletRequest request = new TestHttpSevletRequest(header);
        UploadLogs logs = new UploadLogs();
        assertEquals("remoteHost", LOCALHOST_IP, logs.getIP(request));
        String ip = "12.43.32.12";
        header.put("x-forwarded-for", ip);
        assertEquals("x-forwarded", ip, logs.getIP(request));
    }
    
    public void testRequestParametersCopiedToPageContext() throws Exception {
        
        final Map<String,Object> attr = new HashMap<String,Object>();
        final Map<String,String> param = new HashMap<String,String>();
        HttpServletRequest request = new TestHttpSevletRequest(attr, param);
        
        HttpSession session = new HttpSessionImpl(attr);
        PageContext pg = StatisticsTest.PgCnt.create(session, request);
        
        param.put("ActionsIncludes", ".*All.*");
        param.put("empty", "");
        
        result.preparePageContext(pg, null);
        
        Object value = pg.getAttribute("ActionsIncludes", pg.REQUEST_SCOPE);
        
        assertEquals(".*All.*", value);
        assertNull(pg.getAttribute("empty", pg.REQUEST_SCOPE));
    }
    
    private static class TestHttpSevletRequest implements HttpServletRequest{
        Map<String, Object> attr;
        Map<String, String> param;
        Map<String, String> header;
        
        public TestHttpSevletRequest(Map<String, Object> attr, Map<String, String> param) {
           this.attr = attr;
           this.param = param;
        }
        
        public TestHttpSevletRequest(Map<String, String> header) {
            this.header = header;
        }
        
        public String getAuthType() {
                throw new UnsupportedOperationException("Not supported yet.");
            }

            public Cookie[] getCookies() {
                throw new UnsupportedOperationException("Not supported yet.");
            }

            public long getDateHeader(String arg0) {
                throw new UnsupportedOperationException("Not supported yet.");
            }

            public String getHeader(String arg0) {
                return header.get(arg0);
            }

            public Enumeration getHeaders(String arg0) {
                throw new UnsupportedOperationException("Not supported yet.");
            }

            public Enumeration getHeaderNames() {
                throw new UnsupportedOperationException("Not supported yet.");
            }

            public int getIntHeader(String arg0) {
                throw new UnsupportedOperationException("Not supported yet.");
            }

            public String getMethod() {
                throw new UnsupportedOperationException("Not supported yet.");
            }

            public String getPathInfo() {
                throw new UnsupportedOperationException("Not supported yet.");
            }

            public String getPathTranslated() {
                throw new UnsupportedOperationException("Not supported yet.");
            }

            public String getContextPath() {
                throw new UnsupportedOperationException("Not supported yet.");
            }

            public String getQueryString() {
                throw new UnsupportedOperationException("Not supported yet.");
            }

            public String getRemoteUser() {
                throw new UnsupportedOperationException("Not supported yet.");
            }

            public boolean isUserInRole(String arg0) {
                throw new UnsupportedOperationException("Not supported yet.");
            }

            public Principal getUserPrincipal() {
                throw new UnsupportedOperationException("Not supported yet.");
            }

            public String getRequestedSessionId() {
                throw new UnsupportedOperationException("Not supported yet.");
            }

            public String getRequestURI() {
                throw new UnsupportedOperationException("Not supported yet.");
            }

            public StringBuffer getRequestURL() {
                throw new UnsupportedOperationException("Not supported yet.");
            }

            public String getServletPath() {
                throw new UnsupportedOperationException("Not supported yet.");
            }

            public HttpSession getSession(boolean arg0) {
                throw new UnsupportedOperationException("Not supported yet.");
            }

            public HttpSession getSession() {
                throw new UnsupportedOperationException("Not supported yet.");
            }

            public boolean isRequestedSessionIdValid() {
                throw new UnsupportedOperationException("Not supported yet.");
            }

            public boolean isRequestedSessionIdFromCookie() {
                throw new UnsupportedOperationException("Not supported yet.");
            }

            public boolean isRequestedSessionIdFromURL() {
                throw new UnsupportedOperationException("Not supported yet.");
            }

            public boolean isRequestedSessionIdFromUrl() {
                throw new UnsupportedOperationException("Not supported yet.");
            }

            public Object getAttribute(String arg0) {
                return attr.get(arg0);
            }

            public Enumeration getAttributeNames() {
                throw new UnsupportedOperationException("Not supported yet.");
            }

            public String getCharacterEncoding() {
                throw new UnsupportedOperationException("Not supported yet.");
            }

            public void setCharacterEncoding(String arg0) throws UnsupportedEncodingException {
                throw new UnsupportedOperationException("Not supported yet.");
            }

            public int getContentLength() {
                throw new UnsupportedOperationException("Not supported yet.");
            }

            public String getContentType() {
                throw new UnsupportedOperationException("Not supported yet.");
            }

            public ServletInputStream getInputStream() throws IOException {
                throw new UnsupportedOperationException("Not supported yet.");
            }

            public String getParameter(String arg0) {
                return param.get(arg0);
            }

            public Enumeration getParameterNames() {
                return Collections.enumeration(param.keySet());
            }

            public String[] getParameterValues(String arg0) {
                return param.values().toArray(new String[0]);
            }

            public Map getParameterMap() {
                return param;
            }

            public String getProtocol() {
                throw new UnsupportedOperationException("Not supported yet.");
            }

            public String getScheme() {
                throw new UnsupportedOperationException("Not supported yet.");
            }

            public String getServerName() {
                throw new UnsupportedOperationException("Not supported yet.");
            }

            public int getServerPort() {
                throw new UnsupportedOperationException("Not supported yet.");
            }

            public BufferedReader getReader() throws IOException {
                throw new UnsupportedOperationException("Not supported yet.");
            }

            public String getRemoteAddr() {
                throw new UnsupportedOperationException("Not supported yet.");
            }

            public String getRemoteHost() {
                return LOCALHOST_IP;
            }

            public void setAttribute(String arg0, Object arg1) {
                throw new UnsupportedOperationException("Not supported yet.");
            }

            public void removeAttribute(String arg0) {
                throw new UnsupportedOperationException("Not supported yet.");
            }

            public Locale getLocale() {
                throw new UnsupportedOperationException("Not supported yet.");
            }

            public Enumeration getLocales() {
                throw new UnsupportedOperationException("Not supported yet.");
            }

            public boolean isSecure() {
                throw new UnsupportedOperationException("Not supported yet.");
            }

            public RequestDispatcher getRequestDispatcher(String arg0) {
                throw new UnsupportedOperationException("Not supported yet.");
            }

            public String getRealPath(String arg0) {
                throw new UnsupportedOperationException("Not supported yet.");
            }

            public int getRemotePort() {
                throw new UnsupportedOperationException("Not supported yet.");
            }

            public String getLocalName() {
                throw new UnsupportedOperationException("Not supported yet.");
            }

            public String getLocalAddr() {
                throw new UnsupportedOperationException("Not supported yet.");
            }

            public int getLocalPort() {
                throw new UnsupportedOperationException("Not supported yet.");
            }
            
    }
    
    public static PageContext createPageContext() {
        final Map<String,Object> attr = new HashMap<String,Object>();
        final Map<String,String> param = new HashMap<String,String>();
        HttpServletRequest request = new TestHttpSevletRequest(attr, param);
        
        HttpSession session = new HttpSessionImpl(attr);
        return PgCnt.create(session, request);
    }
    
    public static LogsManager createManager(File f) {
        return LogsManager.createManager(f);
    }
    
    @SuppressWarnings("unchecked")
    private static Map<InputGesture,Integer> getGlobalCounts(LogsManager result) throws InterruptedException, ExecutionException {
        PageContext page = createPageContext();
        result.preparePageContext(page, null);
        return (Map<InputGesture, Integer>) page.getAttribute("globalGestures");
    }
    @SuppressWarnings("unchecked")
    private static Map<InputGesture,Integer> getUserCounts(LogsManager result, String id) throws InterruptedException, ExecutionException {
        PageContext page = createPageContext();
        result.preparePageContext(page, id, null);
        return (Map<InputGesture, Integer>) page.getAttribute("userGestures");
    }
    @SuppressWarnings("unchecked")
    private static Map<InputGesture,Integer> getLastSessionCounts(LogsManager result, String id) throws InterruptedException, ExecutionException {
        PageContext page = createPageContext();
        result.preparePageContext(page, id, null);
        return (Map<InputGesture, Integer>) page.getAttribute("lastGestures");
    }

    private static class HttpSessionImpl implements HttpSession {

        private final Map<String, Object> attr;

        public HttpSessionImpl(Map<String, Object> attr) {
            this.attr = attr;
        }

        public long getCreationTime() {
            throw new UnsupportedOperationException("Not supported yet.");
        }

        public String getId() {
            throw new UnsupportedOperationException("Not supported yet.");
        }

        public long getLastAccessedTime() {
            throw new UnsupportedOperationException("Not supported yet.");
        }

        public ServletContext getServletContext() {
            throw new UnsupportedOperationException("Not supported yet.");
        }

        public void setMaxInactiveInterval(int arg0) {
            throw new UnsupportedOperationException("Not supported yet.");
        }

        public int getMaxInactiveInterval() {
            throw new UnsupportedOperationException("Not supported yet.");
        }

        public HttpSessionContext getSessionContext() {
            throw new UnsupportedOperationException("Not supported yet.");
        }

        public Object getAttribute(String arg0) {
            return attr.get(arg0);
        }

        public Object getValue(String arg0) {
            throw new UnsupportedOperationException("Not supported yet.");
        }

        public Enumeration getAttributeNames() {
            return Collections.enumeration(attr.keySet());
        }

        public String[] getValueNames() {
            throw new UnsupportedOperationException("Not supported yet.");
        }

        public void setAttribute(String arg0, Object arg1) {
            throw new UnsupportedOperationException("Not supported yet.");
        }

        public void putValue(String arg0, Object arg1) {
            throw new UnsupportedOperationException("Not supported yet.");
        }

        public void removeAttribute(String arg0) {
            throw new UnsupportedOperationException("Not supported yet.");
        }

        public void removeValue(String arg0) {
            throw new UnsupportedOperationException("Not supported yet.");
        }

        public void invalidate() {
            throw new UnsupportedOperationException("Not supported yet.");
        }

        public boolean isNew() {
            throw new UnsupportedOperationException("Not supported yet.");
        }
    }
}
