/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 *
 * Copyright 1997-2007 Sun Microsystems, Inc. All rights reserved.
 *
 * The contents of this file are subject to the terms of either the GNU
 * General Public License Version 2 only ("GPL") or the Common
 * Development and Distribution License("CDDL") (collectively, the
 * "License"). You may not use this file except in compliance with the
 * License. You can obtain a copy of the License at
 * http://www.netbeans.org/cddl-gplv2.html
 * or nbbuild/licenses/CDDL-GPL-2-CP. See the License for the
 * specific language governing permissions and limitations under the
 * License.  When distributing the software, include this License Header
 * Notice in each file and include the License file at
 * nbbuild/licenses/CDDL-GPL-2-CP.  Sun designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Sun in the GPL Version 2 section of the License file that
 * accompanied this code. If applicable, add the following below the
 * License Header, with the fields enclosed by brackets [] replaced by
 * your own identifying information:
 * "Portions Copyrighted [year] [name of copyright owner]"
 *
 * Contributor(s):
 *
 * The Original Software is NetBeans. The Initial Developer of the Original
 * Software is Sun Microsystems, Inc. Portions Copyright 1997-2006 Sun
 * Microsystems, Inc. All Rights Reserved.
 *
 * If you wish your version of this file to be governed by only the CDDL
 * or only the GPL Version 2, indicate your decision by adding
 * "[Contributor] elects to include this software in this distribution
 * under the [CDDL or GPL Version 2] license." If you do not indicate a
 * single choice of license, a recipient has the option to distribute
 * your version of this file under either the CDDL, the GPL Version 2 or
 * to extend the choice of license to its licensees as provided above.
 * However, if you add GPL Version 2 code and therefore, elected the GPL
 * Version 2 license, then the option applies only if the new code is
 * made subject to such option by the copyright holder.
 */
package org.netbeans.mdr.persistence.btreeimpl.btreeindex;

import org.netbeans.mdr.persistence.*;
import java.util.*;
import java.io.*;

/**
 * Iterator over a BtreeCollection, which is the set of values contained in
 * a SinglevaluedBtree.
 *
 * @author	Dana Bergen
 * @version	1.0
 */
public class BtreeIterator extends Object implements Iterator {

    protected Btree btree;
    protected SearchResult current;
    protected BtreePageSource pageSource;
    protected int modCount;

    BtreeIterator(Btree btree) {
        try {
	    this.btree = btree;
	    pageSource = btree.pageSource;
	    btree.beginRead();
	    modCount = btree.modCount;
	    current = btree.getFirst();
	    if (current.entryNum >= 0) {
	        current.entryNum--;
	    }
	} catch (StorageException e) {
	    throw new RuntimeStorageException(e);
	} finally {
	    btree.endRead();
	}
    }

    /**
     * Tests whether there are any more elements to return
     *
     * @return	true if a call to next() would succeed
     */
    public boolean hasNext() {

	try {
	    btree.beginRead();
	    checkModCount();
	    if (!BtreePage.hasNext(null, current)) {
	        pageSource.unpinPage(current.page);
		current.page = null;
		return false;
	    } else {
	        return true;
	    }
	} catch (StorageException e) {
	    throw new RuntimeStorageException(e);
	} finally {
	    btree.endRead();
	}
    }

    /**
     * Gets the next value in the btree.  If this is on a different page
     * from the previous value, the old page will be unpinned.
     * If there are no more records, the last page will be unpinned.
     *
     * @return	The next value in the btree
     *
     * @exception	NoSuchElementException	If there was any error or if
     * 						there are no more records
     */
    public Object next() throws NoSuchElementException {

	BtreePage	oldPage;
	
	if (current.page == null) {
	    throw new NoSuchElementException();
	}
	try {
	    btree.beginRead();
	    checkModCount();
	    oldPage = current.page;
	    BtreePage.getNext(null, current);
	    if (current.page != oldPage) {
	        pageSource.unpinPage(oldPage);
	    }
	    if (current.matched) {
		return getCurrentItem();
	    } else {
	        throw new NoSuchElementException();
	    }
	} catch (StorageException e) {
	    throw new RuntimeStorageException(e);
	} finally {
	    btree.endRead();
	}
    }

    /**
     * Returns the data value at the current position
     */
    Object getCurrentItem() throws StorageException {
	return btree.dataInfo.fromBuffer(
				current.page.getData(current.entryNum));
    }

    void checkModCount() {
	if (btree.modCount > modCount) {
	    throw new ConcurrentModificationException("Index " + btree.getName()
		    + " has been modified since iterator was created.");
	}
    }

    /*
     * Unpin the last page.
     */
    protected void finalize() {
	if (current.page != null) {
            pageSource.unpinPage(current.page);
	    current.page = null;
	}
    }

    /**
     * This is not supported.
     *
     * @exception	UnsupportedOperationException	Always thrown.
     */
    public void remove() throws UnsupportedOperationException {
        throw new UnsupportedOperationException();
    }
}
