/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 *
 * Copyright 1997-2007 Sun Microsystems, Inc. All rights reserved.
 *
 * The contents of this file are subject to the terms of either the GNU
 * General Public License Version 2 only ("GPL") or the Common
 * Development and Distribution License("CDDL") (collectively, the
 * "License"). You may not use this file except in compliance with the
 * License. You can obtain a copy of the License at
 * http://www.netbeans.org/cddl-gplv2.html
 * or nbbuild/licenses/CDDL-GPL-2-CP. See the License for the
 * specific language governing permissions and limitations under the
 * License.  When distributing the software, include this License Header
 * Notice in each file and include the License file at
 * nbbuild/licenses/CDDL-GPL-2-CP.  Sun designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Sun in the GPL Version 2 section of the License file that
 * accompanied this code. If applicable, add the following below the
 * License Header, with the fields enclosed by brackets [] replaced by
 * your own identifying information:
 * "Portions Copyrighted [year] [name of copyright owner]"
 *
 * Contributor(s):
 */
package org.foo.wavutils;


import java.io.IOException;

import javax.sound.sampled.AudioFormat;
import javax.sound.sampled.AudioSystem;
import javax.sound.sampled.DataLine;
import javax.sound.sampled.LineUnavailableException;
import javax.sound.sampled.SourceDataLine;

import org.openide.ErrorManager;
import org.openide.util.NbBundle;
import org.openide.util.RequestProcessor;

/**
 * Writes the audio data from an AudioUI to a JavaSound outpout line.
 *
 * @author Rich Unger
 */
public class Player implements Runnable {

    public Player( AudioUI editor )
    {
        audioUI = editor;
    }

    public void start() {
        task = RequestProcessor.getDefault().post(this);
        task.addTaskListener(audioUI);
    }
    
    public void pause()
    {
        if( task != null )
        {
            if( dataLine.isActive() )
            {
                dataLine.stop();
            }
            else
            {
                dataLine.start();
            }
        }
    }

    public void stop() {
        task = null;
    }
    
    private void shutDown()
    {
        SamplingGraph graph = audioUI.getSamplingGraph();
        
        graph.setPlaybackLine(null);

        if (task != null) {
            task = null;
            graph.stop();
        } 
    }

    public void run() {
        WrappedAudioInputStream pcmStream = audioUI.getAudioInputStream();

        // make sure we have something to play
        if (pcmStream == null) {
            ErrorManager.getDefault().log(ErrorManager.ERROR,
                    NbBundle.getMessage(Player.class, "NoAudioDataToLoad"));
            shutDown();
            return;
        }
        
        // if there's a selection, we'll just play that part.
        AudioUI.Selection curSelection = AudioUI.Selection.getSelection(audioUI);

        AudioFormat pcmFormat = pcmStream.getFormat();
                    
        // define the required attributes for our line, 
        // and make sure a compatible line is supported.

        DataLine.Info info = new DataLine.Info(SourceDataLine.class, 
            pcmFormat);
        if (!AudioSystem.isLineSupported(info)) {
            ErrorManager.getDefault().log(ErrorManager.ERROR, 
                    "Line matching " + info + " not supported.");
            shutDown();
            return;
        }

        // get and open the source data line for playback.

        try {
            dataLine = (SourceDataLine) AudioSystem.getLine(info);
            dataLine.open(pcmFormat, BUFSIZE);
        } catch (LineUnavailableException ex) {
            ErrorManager em = ErrorManager.getDefault();
            em.notify(em.annotate(ex, "Unable to open the line: " + ex.getMessage()));
            shutDown();
            return;
        }

        // play back the captured audio data

        int frameSizeInBytes = pcmFormat.getFrameSize();
        int bufferLengthInFrames = dataLine.getBufferSize() / 8;
        int bufferLengthInBytes = bufferLengthInFrames * frameSizeInBytes;
        byte[] data = new byte[bufferLengthInBytes];
        int numBytesRead = 0;
        
        // if there's a selection, just read that
        int iTotalBytesRemaining;
        if( curSelection == null )
        {
            iTotalBytesRemaining = pcmStream.getLength();
        }
        else
        {
            try
            {
                pcmStream.skip( curSelection.offset );
            }
            catch( IOException ex )
            {
                ErrorManager em = ErrorManager.getDefault();
                em.notify(em.annotate(ex, 
                        NbBundle.getMessage(Player.class, "PlaybackError") + ex.getMessage()));
                shutDown();
            }
            iTotalBytesRemaining = curSelection.length;
        }

        // link the line to the sampling graph
        audioUI.getSamplingGraph().setPlaybackLine(dataLine);
        
        // start the source data line
        dataLine.start();

        while (task != null) {
            try {
                if ((numBytesRead = pcmStream.read( data, 0, 
                     Math.min(iTotalBytesRemaining, data.length) ) ) <= 0) {
                    break;
                }
                int numBytesRemaining = numBytesRead;
                while (numBytesRemaining > 0 ) {
                    numBytesRemaining -= dataLine.write(data, 0, numBytesRemaining);
                }
                iTotalBytesRemaining -= numBytesRead;
            } catch (Exception ex) {
                ErrorManager em = ErrorManager.getDefault();
                em.notify(em.annotate(ex,
                        NbBundle.getMessage(Player.class, "PlaybackError") + ex.getMessage()));
                shutDown();
                break;
            }
        }
        // we reached the end of the stream.  reset the stream, 
        // let the data play out, then stop and close the line.
        String sShutDownString = null;
        try
        {
            pcmStream.reset();
        }
        catch( IOException ex )
        {
            ErrorManager em = ErrorManager.getDefault();
            sShutDownString = NbBundle.getMessage(
                Player.class, "AudioStreamResetError") + ex.getMessage();
            em.notify(em.annotate(ex, sShutDownString));
        }
        
        if (task != null) {
            dataLine.drain();
        }
        dataLine.stop();
        dataLine.close();
        dataLine = null;
        shutDown();
    }

    /** The line to which audio data is played */
    private SourceDataLine dataLine;
    
    private RequestProcessor.Task task;
    
    private final AudioUI audioUI;
    
    private static final int BUFSIZE = 16000;
} 
