/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 *
 * Copyright 1997-2007 Sun Microsystems, Inc. All rights reserved.
 *
 * The contents of this file are subject to the terms of either the GNU
 * General Public License Version 2 only ("GPL") or the Common
 * Development and Distribution License("CDDL") (collectively, the
 * "License"). You may not use this file except in compliance with the
 * License. You can obtain a copy of the License at
 * http://www.netbeans.org/cddl-gplv2.html
 * or nbbuild/licenses/CDDL-GPL-2-CP. See the License for the
 * specific language governing permissions and limitations under the
 * License.  When distributing the software, include this License Header
 * Notice in each file and include the License file at
 * nbbuild/licenses/CDDL-GPL-2-CP.  Sun designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Sun in the GPL Version 2 section of the License file that
 * accompanied this code. If applicable, add the following below the
 * License Header, with the fields enclosed by brackets [] replaced by
 * your own identifying information:
 * "Portions Copyrighted [year] [name of copyright owner]"
 *
 * Contributor(s):
 */

package org.foo.wavutils;
/*
 * An AudioInputStream that wraps another stream, and can represent
 * a subset of the enclosed stream (for cropping purposes).
 *
 * When the stream is read, it begins at offset and reads only length
 * bytes of data, even if the enclosed stream may be longer.  This provides
 * a convenient way to take an audio stream and crop it without having to
 * copy the underlying buffer.
 *
 * @author Rich Unger
 */

import javax.sound.sampled.*;
import java.io.*;

public class WrappedAudioInputStream extends AudioInputStream
{
    public WrappedAudioInputStream(AudioInputStream is, int length)
    {
        super(is, is.getFormat(), length); // this doesn't do anything,
                                           // as we'll be reading
                                           // from the wrapped stream
        wrappedStream = is;
        setLength(length);
    }

    public WrappedAudioInputStream(AudioInputStream is)
    {
        super(is, is.getFormat(), is.getFrameLength()); // this doesn't do
                                                        // anything, as we'll
                                                        // be reading from
                                                        // the wrapped stream
        wrappedStream = is;
        setLength( (int)is.getFrameLength() * is.getFormat().getFrameSize() );
    }
    
    public void toFile(File file, AudioFormat format, AudioFileFormat.Type fileType) throws IOException
    {
        // convert to mu-law
        AudioInputStream ulawStream = AudioSystem.getAudioInputStream(format, this);

        OutputStream os = new FileOutputStream(file);

        try
        {
            AudioSystem.write(ulawStream, fileType, os);
        }
        finally
        {
            if (os != null)
            {
                os.close();
            }
            this.reset();
        }
    }


    /** Create a WrappedAudioInputStream from the given File. */
    public static WrappedAudioInputStream createFromFile(File promptFile)
    throws IOException, UnsupportedAudioFileException {
        if( promptFile == null ) {
            return null;
        }

        AudioInputStream ulawStream = null;
        try {
            // the input stream on the prompt file
            ulawStream = AudioSystem.getAudioInputStream( promptFile );

            // convert the stream to PCM
            if( !AudioSystem.isConversionSupported(
                 Formats.INTERNAL_FORMAT, ulawStream.getFormat() ) )
            {
                throw new UnsupportedAudioFileException(
                    "Cannot convert file of format " + ulawStream.getFormat() +
                    " to " + Formats.INTERNAL_FORMAT);
            }
            return createFromUlawStream(ulawStream);
        } finally {
            // close the file (release file locks, etc)
            if( ulawStream != null ) {
                try {
                    ulawStream.close();
                } catch( IOException ex ) {
                    // print stack trace to ide.log
                    ex.printStackTrace();
                }
            }
        }
    }

    /**
     * Convert the given mu-law stream to PCM, and return as a WrappedAudioInputStream.
     * Due to a bug in sun's javasound implementation, we must
     * read in the whole stream now, and deal with the audio data
     * post-conversion, rather than letting the conversion happen
     * lazily.  Otherwise, the cropper will get conflicting information
     * about the sample size.
     */
    public static WrappedAudioInputStream createFromUlawStream(AudioInputStream ulawStream)
    throws IOException {
        AudioInputStream pcmStream = AudioSystem.getAudioInputStream(
                Formats.INTERNAL_FORMAT, ulawStream);

        // read the whole stream into memory before doing any conversions,
        // so we don't lock the file
        byte[] payload = new byte[(int)pcmStream.getFrameLength() *
                                    Formats.INTERNAL_FORMAT.getFrameSize()];
        pcmStream.read(payload);
        pcmStream = new AudioInputStream( new ByteArrayInputStream(payload),
                                            Formats.INTERNAL_FORMAT,
                                            pcmStream.getFrameLength() );

        return new WrappedAudioInputStream( pcmStream );
    }


    public WrappedAudioInputStream convert( AudioFormat newFormat )
    {
        AudioFormat oldFormat = getFormat();

        if( !AudioSystem.isConversionSupported( newFormat, oldFormat ) )
        {
            throw new IllegalArgumentException("Cannot convert file of format " +
                                  oldFormat + " to " + newFormat);
        }
        AudioInputStream newStream = AudioSystem.getAudioInputStream(
                        newFormat, this);

        int iLength = getLength() / oldFormat.getFrameSize()
                                  * newFormat.getFrameSize();

        return new WrappedAudioInputStream( newStream, iLength );
    }

    public void crop( int offset, int length ) throws IOException
    {
        skip( offset );
        setLength(length);
    }

    public WrappedAudioInputStream normalize() throws IOException
    {
        int highSample = -1;
        int sample;

        // find the highest peak
        for( int i=0; i < getFrameLength(); i++ )
        {
            sample = Math.abs(readSample());
            highSample = sample > highSample ? sample : highSample;
        }

        // go back to the start of the stream
        reset();

        double ratio = (double) Short.MAX_VALUE / (double) highSample;

        ByteArrayOutputStream bytesOut = new ByteArrayOutputStream( getLength() );
        DataOutputStream out = new DataOutputStream( bytesOut );

        for( int i=0; i < getFrameLength(); i++ )
        {
            sample = readSample();
            out.writeShort( (short) (ratio * sample) );
        }

        AudioInputStream audioInputStream = new AudioInputStream(
            new ByteArrayInputStream( bytesOut.toByteArray() ),
            Formats.INTERNAL_FORMAT,
            getFrameLength() );

        return new WrappedAudioInputStream( audioInputStream );
    }

    final public void setLength(int i)
    {
        length = i;
        mark(i);
    }

    public int getLength()
    {
        return length;
    }

    public AudioFormat getFormat()
    {
        return wrappedStream.getFormat();
    }

    public long getFrameLength()
    {
        return length / getFormat().getFrameSize();
    }

    public int read() throws IOException
    {
        int n = wrappedStream.read();
        if( n != -1 )
        {
            offset++;
        }
        return n;
    }

    public int read( byte[] buf ) throws IOException
    {
        return read( buf, 0, buf.length );
    }

    public int read( byte[] buf, int off, int len ) throws IOException
    {
        // the "+1" is a hackish workaround for the fact
        // that audio input streams have a nasty habit
        // of leaving one byte on the stream, even for 8-bit audio
        if( offset + len + 1 >= length )
        {
            int n = wrappedStream.read(buf, off, length - offset);

            if( n != -1 )
            {
                offset += n;
            }
            return n;
        }
        else
        {
            int n = wrappedStream.read(buf, off, len);

            if( n != -1 )
            {
                offset += n;
            }
            return n;
        }
    }

    /** reads a single 16-bit, big-endian sample as a short */
    public short readSample() throws IOException
    {
        byte[] buf = new byte[2];
        read(buf);
        return (short)( ( (short)buf[0] << 8 ) + buf[1] );
    }

    public long skip( long n ) throws IOException
    {
        long retval = wrappedStream.skip(n);
        offset += (int)retval;

        return retval;
    }

    public int available() throws IOException
    {
        return length - offset;
    }

    public void close() throws IOException
    {
        wrappedStream.close();
        super.close();
    }

    final public void mark( int readLimit )
    {
        wrappedStream.mark(readLimit);
        offset = 0;
    }

    public void reset() throws IOException
    {
        wrappedStream.reset();
        offset = 0;
    }

    public boolean markSupported()
    {
        return wrappedStream.markSupported();
    }

    public AudioInputStream wrappedStream;
    private int length;
    public int offset;

}
