/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 *
 * Copyright 1997-2007 Sun Microsystems, Inc. All rights reserved.
 *
 * The contents of this file are subject to the terms of either the GNU
 * General Public License Version 2 only ("GPL") or the Common
 * Development and Distribution License("CDDL") (collectively, the
 * "License"). You may not use this file except in compliance with the
 * License. You can obtain a copy of the License at
 * http://www.netbeans.org/cddl-gplv2.html
 * or nbbuild/licenses/CDDL-GPL-2-CP. See the License for the
 * specific language governing permissions and limitations under the
 * License.  When distributing the software, include this License Header
 * Notice in each file and include the License file at
 * nbbuild/licenses/CDDL-GPL-2-CP.  Sun designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Sun in the GPL Version 2 section of the License file that
 * accompanied this code. If applicable, add the following below the
 * License Header, with the fields enclosed by brackets [] replaced by
 * your own identifying information:
 * "Portions Copyrighted [year] [name of copyright owner]"
 *
 * Contributor(s):
 *
 * The Original Software is NetBeans. The Initial Developer of the Original
 * Software is Sun Microsystems, Inc. Portions Copyright 1997-2007 Sun
 * Microsystems, Inc. All Rights Reserved.
 *
 * If you wish your version of this file to be governed by only the CDDL
 * or only the GPL Version 2, indicate your decision by adding
 * "[Contributor] elects to include this software in this distribution
 * under the [CDDL or GPL Version 2] license." If you do not indicate a
 * single choice of license, a recipient has the option to distribute
 * your version of this file under either the CDDL, the GPL Version 2 or
 * to extend the choice of license to its licensees as provided above.
 * However, if you add GPL Version 2 code and therefore, elected the GPL
 * Version 2 license, then the option applies only if the new code is
 * made subject to such option by the copyright holder.
 */
package org.netbeans.examples.customexplorer.gfx;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.datatransfer.Transferable;
import java.awt.datatransfer.UnsupportedFlavorException;
import java.awt.dnd.DropTarget;
import java.awt.dnd.DropTargetDropEvent;
import java.awt.event.InputEvent;
import java.awt.event.KeyEvent;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.beans.PropertyVetoException;
import java.io.IOException;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import javax.swing.Action;
import javax.swing.JPopupMenu;
import javax.swing.JScrollPane;
import javax.swing.KeyStroke;
import javax.swing.MenuElement;
import org.openide.explorer.ExplorerManager;
import org.openide.nodes.Node;
import org.openide.nodes.NodeOp;
import org.openide.util.Exceptions;
import org.openide.util.Utilities;
import org.netbeans.examples.customexplorer.CustomRectangle;
import org.netbeans.examples.customexplorer.ShapesBusinessObject;
import org.netbeans.examples.customexplorer.ShapesContentListener;
/**
 *
 * @author twheeler
 */
public class ShapePanel extends JScrollPane implements MouseListener {

    private Set<Rectangle> rects;
    private Map<CustomRectangle, Node> rectNodeMap;
    private Rectangle selectedRect;
    
    private transient Node root;    
    private transient ExplorerManager manager;
    private static final BasicStroke NORMAL_STROKE = new BasicStroke(2);
    private static final BasicStroke SELECTED_STROKE = new BasicStroke(5);
   
    public ShapePanel() {
        addMouseListener(this);
        rects = new HashSet<Rectangle>();
        rectNodeMap = new HashMap<CustomRectangle, Node>();
        
        DropTarget dt = new MyDropTarget();
        this.setDropTarget(dt);
    }
    
    private class MyDropTarget extends DropTarget {
        @Override
        public synchronized void drop(DropTargetDropEvent dtde) {
            try {
                Transferable t = dtde.getTransferable();
                
                // get the instance data - we know what kind it is, or at least what we support
                CustomRectangle cr = (CustomRectangle) t.getTransferData(CustomRectangle.DATA_FLAVOR);
                
                // get the domain object parent, so we can update the model
                ShapesBusinessObject sbo = root.getLookup().lookup(ShapesBusinessObject.class);
                sbo.add(cr);
            } catch (UnsupportedFlavorException ex) {
                Exceptions.printStackTrace(ex);
            } catch (IOException ex) {
                Exceptions.printStackTrace(ex);
            }
        }
    }
    
    // refreshes the data structures used to build the graphical view
    private void buildRectangleToNodeMap() {
        Rectangle selectedRectCopy = selectedRect;
        
        rectNodeMap.clear();
        rects.clear();
        
        Node[] allKids = root.getChildren().getNodes(true);
        for (int i = 0; i < allKids.length; i++) {
            Node node = allKids[i];
            CustomRectangle rect = node.getLookup().lookup(CustomRectangle.class);
            if (rect != null) {
                rects.add(rect.getRectangle());
                rectNodeMap.put(rect, node);
            }
        }
        
        if (rects.contains(selectedRectCopy)) {
            this.setSelectedRectangle(selectedRectCopy);
        } else {
            this.setSelectedRectangle(null);
        }
    }

    @Override
    public void addNotify() {
        super.addNotify();
        manager = ExplorerManager.find(this);
        assert(manager != null);
        
        // NOTE here we are making some big assumptions about the validity of
        // the nodes, the number and type of children and the content of their 
        // lookups.  In real life, you'd want to be more flexible.
        root = manager.getRootContext();
        buildRectangleToNodeMap();
        
        // update the view when the model changes
        ShapesBusinessObject sbo = root.getLookup().lookup(ShapesBusinessObject.class);
        sbo.addShapesContentListener(new ShapesContentListener() {

            public void shapesAdded() {
                System.out.println("shapes were added -- repaint");
                buildRectangleToNodeMap();
                repaint();
            }

            public void shapesRemoved() {
                System.out.println("shapes were removed -- repaint");
                buildRectangleToNodeMap();
                repaint();
            }
            
        });

        repaint(); 
    }
    
    @Override
    public void removeNotify() {
        rectNodeMap.clear();
        rects.clear();
        repaint();
        
        manager = null;
        super.removeNotify();
    }
    
    private void setSelectedRectangle(Rectangle selectedRect) {
        this.selectedRect = selectedRect;
    }

    public void paint(Graphics g) {
        Graphics2D g2d = (Graphics2D) g;
        
        g2d.clearRect(0, 0, getWidth(), getHeight());

        for (Rectangle rectangle : rects) {
            if (rectangle.equals(selectedRect)) {
                g2d.setStroke(SELECTED_STROKE);
                g2d.setColor(Color.RED);
            } else {
                g2d.setStroke(NORMAL_STROKE);
                g2d.setColor(Color.BLUE);
            }
            
            g2d.draw(rectangle);
        }
    }

    public void mouseClicked(MouseEvent e) {
        Rectangle rect = getRectangleForPoint(e.getPoint());
        setSelectedRectangle(rect);
        repaint();
        
        if (rect == null) {
            try {
                // parent node is selected by default, so that we can right
                // click into "empty space" and paste
                manager.setSelectedNodes(new Node[] { root });
                //manager.setSelectedNodes(new Node[0]);
            } catch (PropertyVetoException ex) {
                Exceptions.printStackTrace(ex);
            }
            
            return;
        }
        
        // determine which node that maps to and then select it
        for (CustomRectangle cr : rectNodeMap.keySet()) {
            if (rect.equals(cr.getRectangle())) {
                try {
                    Node nodeForRectangle = rectNodeMap.get(cr);
                    manager.setSelectedNodes(new Node[]{nodeForRectangle});
                    break;
                } catch (PropertyVetoException ex) {
                    Exceptions.printStackTrace(ex);
                }
            }
        }
    }
    
    private void showPopupMenu(MouseEvent evt) {
        //The commented code below will also work, BUT if you are using
        //any delegate actions like PasteAction, those will always be 
        //disabled.
/*        Node[] nodes = manager.getSelectedNodes();

        JPopupMenu popup = nodes == null ? null : NodeOp.findContextMenu(nodes);
        if (popup != null) {
            
            MenuElement[] items = popup.getSubElements();
            if (items != null && items.length > 0) {
                popup.show(this, evt.getX(), evt.getY());
            }
        }
 */
        Node[] n = manager.getSelectedNodes();
        if (n.length == 1) {
            Action[] a = n[0].getActions(true);
            JPopupMenu popup = Utilities.actionsToPopup(a, this);
            popup.show (this, evt.getX(), evt.getY());
        }
    }

    public void mousePressed(MouseEvent e) {
        if (e.isPopupTrigger()) {
            showPopupMenu(e);
        }
    }

    public void mouseReleased(MouseEvent e) {
        if (e.isPopupTrigger()) {
            showPopupMenu(e);
        }
    }

    public void mouseEntered(MouseEvent e) {
        //throw new UnsupportedOperationException("Not supported yet.");
    }

    public void mouseExited(MouseEvent e) {
        //throw new UnsupportedOperationException("Not supported yet.");
    }

    private Rectangle getRectangleForPoint(Point point) {
        Rectangle retval = null;
        
        for (Rectangle rectangle : rects) {
            if (rectangle.contains(point)) {
                retval = rectangle;
                break;
            }            
        }
        return retval;
    }    
}
