/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 *
 * Copyright 1997-2007 Sun Microsystems, Inc. All rights reserved.
 *
 * The contents of this file are subject to the terms of either the GNU
 * General Public License Version 2 only ("GPL") or the Common
 * Development and Distribution License("CDDL") (collectively, the
 * "License"). You may not use this file except in compliance with the
 * License. You can obtain a copy of the License at
 * http://www.netbeans.org/cddl-gplv2.html
 * or nbbuild/licenses/CDDL-GPL-2-CP. See the License for the
 * specific language governing permissions and limitations under the
 * License.  When distributing the software, include this License Header
 * Notice in each file and include the License file at
 * nbbuild/licenses/CDDL-GPL-2-CP.  Sun designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Sun in the GPL Version 2 section of the License file that
 * accompanied this code. If applicable, add the following below the
 * License Header, with the fields enclosed by brackets [] replaced by
 * your own identifying information:
 * "Portions Copyrighted [year] [name of copyright owner]"
 *
 * Contributor(s):
 *
 * The Original Software is NetBeans. The Initial Developer of the Original
 * Software is Sun Microsystems, Inc. Portions Copyright 1997-2006 Sun
 * Microsystems, Inc. All Rights Reserved.
 *
 * If you wish your version of this file to be governed by only the CDDL
 * or only the GPL Version 2, indicate your decision by adding
 * "[Contributor] elects to include this software in this distribution
 * under the [CDDL or GPL Version 2] license." If you do not indicate a
 * single choice of license, a recipient has the option to distribute
 * your version of this file under either the CDDL, the GPL Version 2 or
 * to extend the choice of license to its licensees as provided above.
 * However, if you add GPL Version 2 code and therefore, elected the GPL
 * Version 2 license, then the option applies only if the new code is
 * made subject to such option by the copyright holder.
 */

package org.netbeans.examples.modules.minicomposer;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.io.CharArrayWriter;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.StringReader;
import java.lang.ref.Reference;
import java.lang.ref.WeakReference;
import java.util.HashSet;
import java.util.Set;
import javax.swing.SwingUtilities;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import javax.swing.event.DocumentEvent;
import javax.swing.event.DocumentListener;
import javax.swing.text.BadLocationException;
import javax.swing.text.Document;
import javax.swing.text.StyledDocument;
import org.openide.ErrorManager;
import org.openide.cookies.EditorCookie;
import org.openide.loaders.DataObject;
import org.openide.text.NbDocument;
import org.openide.util.RequestProcessor;
import org.openide.util.Task;
import org.openide.util.WeakListeners;

public class ScoreSupport implements ScoreCookie, Runnable, DocumentListener, PropertyChangeListener {
    private final ErrorManager err;
    private final DataObject obj;
    private final EditorCookie.Observable edit;
    private Task prepareTask = null;
    private final Set listeners = new HashSet();
    private Score score = null;
    private IOException parseException = null;
    private boolean addedEditorListener = false;
    private Reference lastUsedDocument = null; // Reference<Document>
    public ScoreSupport(DataObject obj, EditorCookie.Observable edit) {
        this.obj = obj;
        err = ErrorManager.getDefault().
            getInstance("org.netbeans.examples.modules.minicomposer.ScoreSupport." + // NOI18N
                        obj.getPrimaryFile().getName());
        this.edit = edit;
    }
    public void propertyChange(PropertyChangeEvent ev) {
        if (EditorCookie.Observable.PROP_DOCUMENT.equals(ev.getPropertyName())) {
            err.log("Editor state changed");
            invalidate();
        }
    }
    public synchronized void addChangeListener(ChangeListener l) {
        listeners.add(l);
    }
    public synchronized void removeChangeListener(ChangeListener l) {
        listeners.remove(l);
    }
    protected synchronized void fireChange() {
        final ChangeListener[] ls = (ChangeListener[])listeners.toArray(new ChangeListener[listeners.size()]);
        if (ls.length == 0) return;
        final ChangeEvent ev = new ChangeEvent(this);
        SwingUtilities.invokeLater(new Runnable() {
            public void run() {
                err.log("firing change");
                for (int i = 0; i < ls.length; i++) {
                    ls[i].stateChanged(ev);
                }
            }
        });
    }
    public synchronized Task prepare() {
        if (prepareTask == null) {
            err.log("preparing to parse");
            prepareTask = RequestProcessor.getDefault().post(this);
        }
        return prepareTask;
    }
    public void run() {
        err.log("run");
        if (!obj.isValid()) {
            err.log("object invalidated");
            setScoreAndParseException(null, new FileNotFoundException(obj.getPrimaryFile().getPath()));
            return;
        }
        edit.prepareDocument().waitFinished();
        final Document doc = edit.getDocument();
        if (doc == null) {
            // Should not happen:
            err.log(ErrorManager.WARNING, "WARNING: Doc was null!");
            return;
        }
        if (!addedEditorListener) {
            err.log("adding editor listener");
            addedEditorListener = true;
            edit.addPropertyChangeListener(WeakListeners.propertyChange(this, edit));
        }
        doc.render(new Runnable() {
            public void run() {
                try {
                    setScoreAndParseException(parse(doc), null);
                } catch (IOException ioe) {
                    setScoreAndParseException(score, ioe);
                } catch (BadLocationException ble) {
                    IOException ioe = new IOException(ble.toString());
                    err.annotate(ioe, ble);
                    setScoreAndParseException(score, ioe);
                }
            }
        });
        Document lastDoc = null;
        if (lastUsedDocument != null) {
            lastDoc = (Document)lastUsedDocument.get();
        }
        if (lastDoc != doc) {
            if (lastDoc != null) {
                err.log("removing listener from old document");
                lastDoc.removeDocumentListener(this);
            }
            err.log("adding fresh document listener");
            doc.addDocumentListener(this);
            lastUsedDocument = new WeakReference(doc);
        }
    }
    private synchronized void setScoreAndParseException(Score s, IOException e) {
        if (err.isLoggable(ErrorManager.INFORMATIONAL)) {
            err.log("parsed; exception=" + e + "; score size=" + s.getSize());
        }
        score = s;
        parseException = e;
        fireChange();
    }
    public boolean isValid() {
        return parseException == null;
    }
    public void changedUpdate(DocumentEvent ev) {
    }
    public void insertUpdate(DocumentEvent ev) {
        invalidate();
    }
    public void removeUpdate(DocumentEvent ev) {
        invalidate();
    }
    protected synchronized void invalidate() {
        err.log("invalidated");
        if (prepareTask != null) {
            prepareTask = null;
            fireChange();
        }
    }
    public synchronized void setScore(final Score s) throws IOException {
        final Score oldScore = score;
        if (s.equals(oldScore)) {
            return;
        }
        err.log("setScore");
        prepareTask = Task.EMPTY;
        score = s;
        parseException = null;
        final StyledDocument doc = edit.openDocument();
        final BadLocationException[] e = new BadLocationException[] {null};
        try {
            NbDocument.runAtomic(doc, new Runnable() {
                public void run() {
                    doc.removeDocumentListener(ScoreSupport.this);
                    err.log("removed doc listener");
                    try {
                        generate(s, oldScore, doc);
                    } catch (BadLocationException ble) {
                        e[0] = ble;
                    } finally {
                        err.log("readded doc listener");
                        doc.addDocumentListener(ScoreSupport.this);
                    }
                }
            });
            if (e[0] != null) throw e[0];
        } catch (BadLocationException ble) {
            IOException ioe = new IOException(ble.toString());
            err.annotate(ioe, ble);
            throw ioe;
        }
        fireChange();
    }
    public Score getScore() throws IOException {
        prepare().waitFinished();
        synchronized (this) {
            if (score != null && (parseException == null || obj.isModified())) {
                return score;
            } else {
                if (parseException == null) {
                    // Should not happen:
                    throw new IOException("parse did not finish as expected");
                }
                throw parseException;
            }
        }
    }
    protected Score parse(Document doc) throws IOException, BadLocationException {
        String text = doc.getText(0, doc.getLength());
        return Score.parse(new StringReader(text));
    }
    protected void generate(Score s, Score oldScore, Document doc) throws BadLocationException {
        CharArrayWriter wr = new CharArrayWriter();
        try {
            Score.generate(s, wr);
        } catch (IOException ioe) {
            // Should not happen.
            err.notify(ioe);
            return;
        }
        doc.remove(0, doc.getLength());
        doc.insertString(0, wr.toString(), null);
    }
}
