/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 *
 * Copyright 1997-2007 Sun Microsystems, Inc. All rights reserved.
 *
 * The contents of this file are subject to the terms of either the GNU
 * General Public License Version 2 only ("GPL") or the Common
 * Development and Distribution License("CDDL") (collectively, the
 * "License"). You may not use this file except in compliance with the
 * License. You can obtain a copy of the License at
 * http://www.netbeans.org/cddl-gplv2.html
 * or nbbuild/licenses/CDDL-GPL-2-CP. See the License for the
 * specific language governing permissions and limitations under the
 * License.  When distributing the software, include this License Header
 * Notice in each file and include the License file at
 * nbbuild/licenses/CDDL-GPL-2-CP.  Sun designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Sun in the GPL Version 2 section of the License file that
 * accompanied this code. If applicable, add the following below the
 * License Header, with the fields enclosed by brackets [] replaced by
 * your own identifying information:
 * "Portions Copyrighted [year] [name of copyright owner]"
 *
 * Contributor(s):
 *
 * The Original Software is NetBeans. The Initial Developer of the Original
 * Software is Sun Microsystems, Inc. Portions Copyright 1997-2006 Sun
 * Microsystems, Inc. All Rights Reserved.
 *
 * If you wish your version of this file to be governed by only the CDDL
 * or only the GPL Version 2, indicate your decision by adding
 * "[Contributor] elects to include this software in this distribution
 * under the [CDDL or GPL Version 2] license." If you do not indicate a
 * single choice of license, a recipient has the option to distribute
 * your version of this file under either the CDDL, the GPL Version 2 or
 * to extend the choice of license to its licensees as provided above.
 * However, if you add GPL Version 2 code and therefore, elected the GPL
 * Version 2 license, then the option applies only if the new code is
 * made subject to such option by the copyright holder.
 */

package org.netbeans.modules.vcs.cmdline;

import java.io.*;
import java.util.*;
import java.beans.*;

import org.openide.filesystems.FileSystem;

import org.netbeans.modules.vcscore.VcsConfigVariable;
import org.netbeans.modules.vcscore.VcsFileSystem;

import org.netbeans.modules.vcs.advanced.CommandLineVcsFileSystem;

/** CVS filesystem.
 * This class acts as a transfer from CvsFileSystem located in vcscvs module to
 * CommandLineVcsFileSystem with the appropriate CVS profile located in
 * vcsgeneric module.
 * Since vcscvs module was removed from the standard distribution, the
 * filesystems mounted via CvsFileSystem are deserialized and converted
 * to CommandLineVcsFileSystem with CVS profile.
 * 
 * @author Martin Entlicher
 */
//-------------------------------------------
public class CvsFileSystem extends VcsFileSystem implements java.beans.PropertyChangeListener {
    
    //private Debug D = new Debug ("CvsFileSystem", true); // NOI18N
    
    // suppose that CVS server only exists for UNIX/Linux.
    // The same presumption is made in CvsAction.doCommand () method
    final static String PS_SERVER = "/"; // NOI18N
    final static String PS_LOCAL = java.io.File.separator;
    
    private static final String[] VARIABLES_TO_TRANSFER = { "SERVERTYPE", "CVS_SERVER", "CVS_USERNAME",
                                                            "CVS_REPOSITORY", "CVS_EXE", "SHELL",
                                                            VAR_ENVIRONMENT_PREFIX+"CVS_RSH" };

    private String cvsServerType;// = new String(last_cvsServerType);
    private String cvsRoot;// = new String(last_cvsRoot);
    private String cvsServer;// = new String(last_cvsServer);
    private String cvsUser;// = new String(last_cvsUser);
    private String cvsPassword;// = new String(last_cvsPassword);
    private String cvsModule;// = new String(last_cvsModule);
    private String cvsModuleName;// = new String(last_cvsModuleName);
    private String cvsExePath;// = (last_cvsExePath == null) ? null : new String(last_cvsExePath);
    private String cygwinPath;// = (last_cygwinPath == null) ? null : new String(last_cygwinPath);
    private transient boolean doInitialCheckout = false;
    //private boolean useUnixShell = last_useUnixShell;

    private static final String CACHE_FILE_FOLDER = "CVS";
    private static final String CACHE_FILE_NAME = "netbeans.cmd.cache";
    
    private boolean executePreCommit = true;

    //-------------------------------------------
    static final long serialVersionUID =3105697696081480308L;
    public CvsFileSystem() {
    }

    private static final String CONFIG_CVS = "cvs.xml"; // NOI18N
    
    private CommandLineVcsFileSystem createCommandLineVcsFileSystem() {
        String configFileName;
        configFileName = CONFIG_CVS;
        CommandLineVcsFileSystem fs = new CommandLineVcsFileSystem();
        fs.readConfiguration(configFileName);
        fs.setConfigFileName(configFileName);
        Vector vars = fs.getVariables();
        HashMap varsByName = new HashMap();
        for (int i = 0, n = vars.size (); i < n; i++) {
            VcsConfigVariable var = (VcsConfigVariable) vars.get (i);
            varsByName.put (var.getName (), var);
        }
        Set clvfsVarNames = varsByName.keySet();
        Hashtable cvsVariables = getVariablesByName();
        for (int i = 0; i < VARIABLES_TO_TRANSFER.length; i++) {
            String cvsVarName = VARIABLES_TO_TRANSFER[i];
            VcsConfigVariable cvsVar = (VcsConfigVariable) cvsVariables.get(cvsVarName);
            if (cvsVar != null) {
                VcsConfigVariable var = (VcsConfigVariable) varsByName.get(cvsVarName);
                if (var != null) {
                    var.setValue(cvsVar.getValue());
                } else {
                    vars.add(cvsVar);
                }
            } else if ("CVS_USERNAME".equals(cvsVarName)) {
                VcsConfigVariable var = (VcsConfigVariable) varsByName.get(cvsVarName);
                if (var != null) {
                    var.setValue(cvsUser);
                } else {
                    vars.add(new VcsConfigVariable("CVS_USERNAME", "", cvsUser, false, false, false, null));
                }
            }
        }
        // Assure, that for local connection the server and user name will be empty. (See #17944).
        if ("local".equals(cvsServerType)) { // NOI18N
            VcsConfigVariable var = (VcsConfigVariable) varsByName.get("CVS_SERVER"); // NOI18N
            if (var != null) {
                var.setValue(""); // NOI18N
            }
            var = (VcsConfigVariable) varsByName.get("CVS_USERNAME"); // NOI18N
            if (var != null) {
                var.setValue(""); // NOI18N
            }
        }
        fs.setVariables(vars);
        fs.setExpertMode(isExpertMode());
        fs.setAnnotationPattern(getAnnotationPattern());
        fs.setAutoRefresh(getAutoRefresh());
        fs.setCommandNotification(isCommandNotification());
        fs.setOffLine(isOffLine());
        fs.setDebug(getDebug());
        fs.setProcessUnimportantFiles(isProcessUnimportantFiles());
        fs.setHideShadowFiles(isHideShadowFiles());
        return fs;
    }

    /**
     * Get the full file path where cache information should be stored.
     */
    public String getCacheFileName(String path) {
        File cacheDir = getFile(path + File.separator + CACHE_FILE_FOLDER);
        if (!cacheDir.exists() || !cacheDir.isDirectory()) return null;
        else return new File(cacheDir, CACHE_FILE_NAME).getAbsolutePath();
        //return getFile(path + File.separator + CACHE_FILE_FOLDER + File.separator + CACHE_FILE_NAME).getAbsolutePath();
    }
    
    private Hashtable getVariablesByName() {
        return variablesByName;
    }

    public void propertyChange (PropertyChangeEvent evt) {
        if (evt.getPropertyName() != FileSystem.PROP_VALID) return;
        if (isValid()) {
            //D.deb("Filesystem added to the repository, setting refresh time to "+refreshTimeToSet); // NOI18N
            setRefreshTime(refreshTimeToSet);
            //D.deb("calling WARN");
            warnDirectoriesDoNotExists();
            //D.deb("calling WARN finished.");
        } else {
            //D.deb("Filesystem is not valid any more, setting refresh time to 0"); // NOI18N
            setRefreshTime(0);
        }
    }

    private void readObject(ObjectInputStream in) throws
        ClassNotFoundException, IOException, NotActiveException{
        in.defaultReadObject();
        //last_cvsExePath = cvsExePath;
        //last_cvsModule = cvsModule;
        //last_cvsModuleName = cvsModuleName;
        //last_cvsPassword = cvsPassword;
        //last_cvsRoot = cvsRoot;
        //last_cvsServer = cvsServer;
        //last_cvsServerType = cvsServerType;
        //last_cvsUser = cvsUser;
        //last_cygwinPath = cygwinPath;
        /*
        initCommands();
        addKnownStatuses();
        addBadgeIcons();
        setOutputTab();  //MK
        this.setIgnoreListSupport (new CvsIgnoreList());
         */
        setCreateVersioningSystem(true);
    }

    //-------------------------------------------
    private void writeObject(ObjectOutputStream out) throws IOException {
        //D.deb("writeObject() - saving bean"); // NOI18N
        // cache is transient
        out.defaultWriteObject();
    }

    private Object readResolve() throws ObjectStreamException {
        return createCommandLineVcsFileSystem();
    }
}

